from .Entity import Entity


EXTRA_HEADER = """File automatically generated by pymoordyn"""


class Intro(Entity):
    """The introduction section
    """
    def __init__(self):
        """Constructor
        """
        Entity.__init__(self,
                        "MoorDyn Input File",
                        extra_header=EXTRA_HEADER)
